#include "glad/glad.h"
#include "GLFW/glfw3.h"
#include <stdlib.h>
#include <stdio.h>
#include <Psapi.h>

static void error_callback(int error, const char* description)
{
	fprintf(stderr, "Error: %s\n", description);
}

static void key_callback(GLFWwindow* window, int key, int scancode, int action, int mods)
{
	if (key == GLFW_KEY_ESCAPE && action == GLFW_PRESS)
		glfwSetWindowShouldClose(window, GLFW_TRUE);
}


char* loadTextFile(char* path, int* length)
{
	HANDLE hFile = CreateFile(path, FILE_GENERIC_READ, 0, 0, OPEN_EXISTING, 0, 0);
	if (hFile == INVALID_HANDLE_VALUE)
	{
		printf("Unable to load file\n");
		return NULL;
	}
	DWORD fileSize = GetFileSize(hFile, NULL);
	char* fileData = (char*)malloc(fileSize + 1);
	DWORD bt;
	ReadFile(hFile, fileData, fileSize, &bt, NULL);
	fileData[fileSize] = '\0';
	CloseHandle(hFile);
	return fileData;
}

int main(void)
{
	GLFWwindow* window;
	GLuint vertex_shader, fragment_shader, program;
	glfwSetErrorCallback(error_callback);
	if (!glfwInit())
		exit(EXIT_FAILURE);
	glfwWindowHint(GLFW_CONTEXT_VERSION_MAJOR, 2);
	glfwWindowHint(GLFW_CONTEXT_VERSION_MINOR, 0);
	window = glfwCreateWindow(640, 480, "Shader RAM usage test", NULL, NULL);
	if (!window)
	{
		glfwTerminate();
		exit(EXIT_FAILURE);
	}
	glfwSetKeyCallback(window, key_callback);
	glfwMakeContextCurrent(window);
	glfwSwapInterval(1);

	gladLoadGL();

	int vertexShaderLen = 0;
	int fragmentShaderLen = 0;
	char* vertexShaderSrc;
	char* fragmentShaderSrc;

	// example shaders
	// result = amount of RAM allocated by application after shader was compiled and linked 1000 times

	// large shader - result: ~2457MB
	vertexShaderSrc = loadTextFile("shaders\\largeShaderExample_vs_source.txt", &vertexShaderLen);
	fragmentShaderSrc = loadTextFile("shaders\\largeShaderExample_ps_source.txt", &fragmentShaderLen);

	// large shader2 - result: ~2375MB
	// same as shader above, but decreased size of uniform buffers from 4096 to 100
	//vertexShaderSrc = loadTextFile("shaders\\largeShaderExample2_vs_source.txt", &vertexShaderLen);
	//fragmentShaderSrc = loadTextFile("shaders\\largeShaderExample2_ps_source.txt", &fragmentShaderLen);

	// small shader - result: ~103MB
	//vertexShaderSrc = loadTextFile("shaders\\smallShaderExample_vs_source.txt", &vertexShaderLen);
	//fragmentShaderSrc = loadTextFile("shaders\\smallShaderExample_ps_source.txt", &fragmentShaderLen);
	
	// medium shader - result: ~1650MB
	//vertexShaderSrc = loadTextFile("shaders\\mediumShaderNoUniformBuffer_vs_source.txt", &vertexShaderLen);
	//fragmentShaderSrc = loadTextFile("shaders\\mediumShaderNoUniformBuffer_ps_source.txt", &fragmentShaderLen);
	
	// medium shader (mostly float variables) - result: ~1180MB
	//vertexShaderSrc = loadTextFile("shaders\\mediumFloatOnly_vs_source.txt", &vertexShaderLen);
	//fragmentShaderSrc = loadTextFile("shaders\\mediumFloatOnly_ps_source.txt", &fragmentShaderLen);
	
	// medium shader (same as above but with some redundant lines of code manually removed) - result: ~1154MBMB 
	//vertexShaderSrc = loadTextFile("shaders\\mediumFloatOnly2_vs_source.txt", &vertexShaderLen);
	//fragmentShaderSrc = loadTextFile("shaders\\mediumFloatOnly2_ps_source.txt", &fragmentShaderLen);
	
	// medium shader (same as above but with conditional removed from mul_nonIEEE() function) - result: ~343MB (!)
	//vertexShaderSrc = loadTextFile("shaders\\mediumFloatOnly3_vs_source.txt", &vertexShaderLen);
	//fragmentShaderSrc = loadTextFile("shaders\\mediumFloatOnly3_ps_source.txt", &fragmentShaderLen);

	// allocate temporary shader source memory
	char* vertexShaderTempSrc = (char*)malloc(1024 * 1024 * 1);
	char* fragmentShaderTempSrc = (char*)malloc(1024 * 1024 * 1);
	// remember current amount of commited memory
	PROCESS_MEMORY_COUNTERS pmc1;
	GetProcessMemoryInfo(GetCurrentProcess(), &pmc1, sizeof(PROCESS_MEMORY_COUNTERS));
	LONGLONG totalMem1 = pmc1.PagefileUsage;
	// compile shader variations (attach random string at the end to circumvent shader cache)
	unsigned int startTime = GetTickCount();
	for (int i = 0; i < 1000; i++)
	{
		// attach random string
		strcpy(vertexShaderTempSrc, vertexShaderSrc);
		strcpy(fragmentShaderTempSrc, fragmentShaderSrc);
		// attach random string to avoid any internal caching by the graphics driver
		sprintf(vertexShaderTempSrc + strlen(vertexShaderTempSrc), "\r\n// temp random str %08x %08x\r\n", GetTickCount(), i);
		sprintf(fragmentShaderTempSrc + strlen(fragmentShaderTempSrc), "\r\n// temp random str %08x %08x\r\n", GetTickCount(), i);
		// compile
		vertex_shader = glCreateShader(GL_VERTEX_SHADER);
		glShaderSource(vertex_shader, 1, &vertexShaderTempSrc, NULL);
		glCompileShader(vertex_shader);
		fragment_shader = glCreateShader(GL_FRAGMENT_SHADER);
		glShaderSource(fragment_shader, 1, &fragmentShaderTempSrc, NULL);
		glCompileShader(fragment_shader);
		program = glCreateProgram();
		glAttachShader(program, vertex_shader);
		glAttachShader(program, fragment_shader);
		glLinkProgram(program);
		// make sure shader compiled ok
		GLint linkSuccess = 0;
		glGetProgramiv(program, GL_LINK_STATUS, &linkSuccess);
		if (linkSuccess == GL_FALSE)
		{
			printf("glLinkProgram() failed\n");
		}
		printf("%04d/1000\n", i);
		// detach vertex/fragment shader since we no longer need them
		glDetachShader(program, vertex_shader);
		glDetachShader(program, fragment_shader);
		glDeleteShader(vertex_shader);
		glDeleteShader(fragment_shader);
	}
	unsigned int endTime = GetTickCount();
	printf("Compiled all shaders in %dms\n", endTime-startTime);

	PROCESS_MEMORY_COUNTERS pmc2;
	GetProcessMemoryInfo(GetCurrentProcess(), &pmc2, sizeof(PROCESS_MEMORY_COUNTERS));
	LONGLONG totalMem2 = pmc2.PagefileUsage;
	printf("Committed mem: %dMB\n", (int)((totalMem2-totalMem1)/1024ULL/1024ULL));

	while (!glfwWindowShouldClose(window))
	{
		glfwSwapBuffers(window);
		glfwPollEvents();
	}
	glfwDestroyWindow(window);
	glfwTerminate();
	exit(EXIT_SUCCESS);
}